/*
 * RJunglePrediction.h
 *
 *  Created on: 18.12.2008
 *      Author: schwarz
 */

#ifndef RJUNGLEPREDICTION_H_
#define RJUNGLEPREDICTION_H_

/*
 * Includes
 */

#include <iostream>
#include <vector>
#include <ctime>

#include "RJunglePar.h"
#include "RJungleIO.h"
#include "RJungleGen.h"
#include "DataFrame.h"
#include "DataTreeSet.h"
#include "CmpldTree.h"
#include "Helper.h"
#include "treedefs.h"
#include "Prediction.h"

// Prediction stuff
#include <boost/algorithm/string.hpp>
#include <boost/token_iterator.hpp>

template<class T>
class RJunglePrediction {
public:
  RJunglePrediction();
  virtual ~RJunglePrediction();

  static void showPredictionCmpld(
      RJungleIO &io, DataFrame<T> &data, Prediction<T> &pred,
      DataTreeSet &dataSet, uli_t iteration = 0) {

    std::vector<T> classRes;
    std::vector<T> classVec;
    std::vector<T> colVec;
    std::vector<double> accumulateVec;

    uli_t j, m;
    T medianVote;

    if(data.par.probprediction_flag)
      *io.outVerbose << "Writing probability prediction ..." << std::endl;
    else
      *io.outVerbose << "Writing prediction ..." << std::endl;


    for (j = 0; j < pred.nrow; ++j) {
      if (j >= dataSet.getNsmpl())
        throw Exception(ERRORCODE_31);

      // forest
      classRes.clear();
      accumulateVec.clear();

      for (m = 0; m < dataSet.getNtree(); ++m)
        if (dataSet.at(j, m))
          classRes.push_back(pred.pred[j + data.getnrow() * m]);



      if(data.par.probprediction_flag) {

          accumulateVec = accumulateMultiClassProb(classRes, data);

          // write out data
          for(unsigned int i = 0; i < accumulateVec.size(); ++i){
              *io.outPrediction << accumulateVec[i] << "\t";
          }
          *io.outPrediction << std::endl;

      } else {

          medianVote = Helper::getMostFreqProp<T>(
              data.par, classRes, data.getMissingCode(), &data, false, NULL, j);

          if (medianVote != data.getMissingCode())
            *io.outPrediction << (double) medianVote << std::endl;
          else
            *io.outPrediction << "NA" << std::endl;
      }
    }

  }



  // Multiclass prediction functions

  static std::vector <double> accumulateMultiClassProb (
      std::vector<T> &predVec,
      DataFrame<T> &data) {

    std::vector<std::vector <double> > classProbStore;

    for(uli_t i = 0; i < predVec.size(); ++i) {
        int pred = predVec[i];
        classProbStore.push_back(getMultiClassProb(pred, data));
    }

    std::vector<double> accumulateVec (data.par.categories, 0);

    for(uli_t j = 0; j < accumulateVec.size(); ++j){
        for(uli_t i = 0; i < classProbStore.size(); ++i) {
            accumulateVec[j] += classProbStore[i][j];
            //std::cout << classProbStore[i][j] << " ";
        }
        accumulateVec[j] = accumulateVec[j] / data.par.ntree;
    }

    // make vectors clean
    classProbStore.clear();

    return(accumulateVec);
  }



  static bool onlyOne( int &res) {
    std::string predRes01 = Helper::NumberToString(res);
    size_t found = predRes01.find("999");

    if((found != std::string::npos) | (res == 0))
      return(1);
    else
      return(0);
  }

  static std::vector<int> splitInt (int &res) {
    std::vector<int> storeInt;
    while (res > 0)
      {
        int digit = res%10;
        res /= 10;
        storeInt.push_back(digit);
      }
    reverse(storeInt.begin(), storeInt.end());
    return(storeInt);
  }

  static std::vector <std::string> splitStringBy2(std::string s, int groups){
    std::vector<std::string> strs;
    std::vector<int> offset;

   for(size_t i = 0; i < groups; ++i)
     offset.push_back(2);

   boost::offset_separator f(offset.begin(), offset.end());
   typedef boost::token_iterator_generator<boost::offset_separator>::type Iter;
   Iter beg = boost::make_token_iterator<std::string>(s.begin(),s.end(),f);
   Iter end = boost::make_token_iterator<std::string>(s.end(),s.end(),f);
   // The above statement could also have been what is below
   // Iter end;
   for(;beg!=end;++beg){
       strs.push_back(*beg);
   }
   return(strs);
  }

  static double StringToNumberDouble ( const std::string &Text )//Text not by const reference so that the function can be used with a
  {                               //character array as argument
    std::stringstream ss(Text);
    double result;
    return ss >> result ? result : 0;
  }


  static std::vector<double> getMultiClassProb (int &prediction, DataFrame<T> &data){

    // Store vectors
   std::vector<std::string> store;
   std::vector<double> storeDub;

   // check if there has one group a probability of one
   if(onlyOne(prediction)){
       // if prediction is not 0
       if(prediction){

           // get rid of the 999 as indicator
           std::string cleanPredictionString = Helper::remove(Helper::NumberToString(prediction), "999");
           int cleanPrediction = Helper::stringToNumber(cleanPredictionString);


           // Build up empty probability vector
           for(unsigned int i = 0; i < data.par.categories; ++i)
             storeDub.push_back(0);

           // split prediction into vector
           std::vector<int> storeInt = splitInt(cleanPrediction);

           // get the position of the 1
           size_t pos = data.par.categories - storeInt.size() -1;

           // write the 1 into the right position
           storeDub.erase(storeDub.begin()+pos);
           storeDub.insert(storeDub.begin()+pos, 1);
       }else{
           // if prediction is 0 the last probability was 1
           for(unsigned int i = 0; i < data.par.categories-1; ++i)
             storeDub.push_back(0);
           storeDub.push_back(1);
       }
       return(storeDub);
   } else {
       // Prediction to string
       std::string predPrediction = Helper::NumberToString(prediction);

       // solve some problems with 0.01
       if(predPrediction.size() % 2)
         predPrediction.insert(0, "0");

       while(predPrediction.size() < 2*(data.par.categories-1))
         predPrediction.insert(0, "00");

       // Split the string by 2 x groups
       std::vector<std::string> strs;
       strs = splitStringBy2(predPrediction, data.par.categories);

       // add zero point to split number
       for(unsigned int i = 0; i < strs.size(); ++i)
         if(strs[i].size() == 2)
           store.push_back(std::string("0.") + strs[i]);
         else
           store.push_back(std::string("0.0") + strs[i]);

       // convert the string to double
       for(unsigned int i = 0; i < store.size(); ++i)
         storeDub.push_back(StringToNumberDouble(store[i]));

       storeDub.push_back(1 - std::accumulate(storeDub.begin(),storeDub.end(), 0.f));

       return(storeDub);
   }
  }

};

#endif /* RJUNGLEPREDICTION_H_ */
