/*
 *     A subroutine for computing non-central multivariate t probabilities.
 *     This subroutine uses an algorithm (QRSVN) described in the paper
 *     "Comparison of Methods for the Computation of Multivariate
 *         t-Probabilities", by Alan Genz and Frank Bretz
 *         J. Comp. Graph. Stat. 11 (2002), pp. 950-971.
 *
 *          Alan Genz
 *          Department of Mathematics
 *          Washington State University
 *          Pullman, WA 99164-3113
 *          Email : AlanGenz@wsu.edu
 *
 *   Original source available from
 *   http://www.math.wsu.edu/faculty/genz/software/fort77/mvtdstpack.f
 */

#ifndef MVT_H_
#define MVT_H_

#include <gsl/gsl_rng.h>
#include <gsl/gsl_cdf.h>

extern"C" {
int gsl_rng_uniform_fortran_(double *x, gsl_rng **r, int pointerwaste); //gfortran
int gsl_rng_uniform_fortran__(double *x, gsl_rng **r, int pointerwaste); //g77
}

extern"C" {
void mvtdst_(
  int *n, int *nu, double *lower, double *upper,
  int *infin, double *corr, double *delta,
  int *maxpts, double *abseps, double *releps, double *error,
  double *value, int *inform,
  gsl_rng **r
);
}


#endif /* MVT_H_ */
